# file: Flow.py
# author: Bill Ames, Fall 2013
# substantially rewritten by: Bob Muller, Fall 2013
# date: November 10, 2013

# This is harness code for a program that solves the Flow game.
# See e.g., http://moh97.us/flow/
#
from Tkinter import *
import pprint

# Debugging
#
DEBUG = False

# The Board
#
# This version of flow is played on an 6 x 6 board.
#
N = 6

# Shapes
#
# There are 6 playable shapes and 1 unplayble end shape.
#
nShapes = 7    #  == 6 + 1

# Give the shapes numbers and names.
#                         
shapes = range(nShapes)
[Horizontal, Vertical, Q1, Q2, Q3, Q4, Circle] = shapes
shapeNames = ['Horizontal', 'Vertical', 'Q1', 'Q2', 'Q3', 'Q4', 'Circle']

# NB:
#
# Q1 = |    Q2 =   |  Q3 = --+  Q4 = +--
#      +--       --+         |       |
#
def shapeOf(piece): return piece % nShapes
def shapeName(piece): return shapeNames[shapeOf(piece)]

def isEndPiece(piece): return shapeOf(piece) == Circle

# Colors
#
# For now, we have 9 colors: blue, cyan, green, magenta, orange, 
# pink, red, white and yellow.
#
nColors = 9
colors = range(nColors)
colorNames = ['blue', 'cyan', 'green', 'magenta', 'orange', 'pink', 'red', 'white', 'yellow']

def colorOf(piece): return piece / nShapes
def colorName(piece): return colorNames[colorOf(piece)]

def sameColor(piece1, piece2): return colorOf(piece1) == colorOf(piece2)

# Pieces
#
pieces = range(nShapes * nColors)

# Circles are end pieces. They are specified by the user so they
# are unplayable by the program.
#
playablePieces = [ p for p in pieces if not isEndPiece(p)]

# Two special pieces for empty cells and an outer wall.
#
[Empty, Wall] = [len(pieces), len(pieces) + 1]

# pieceName : piece -> string
#
# The call pieceName(piece) returns the name of the piece.
#
def pieceName(piece):
    if piece == Wall:
        return 'Wall'
    elif piece == Empty:
        return 'Empty'
    else:
        return colorName(piece) + shapeName(piece)

# isEmpty : piece -> boolean
#
def isEmpty(piece): return piece == Empty

# placePiece : piece * board * row * col -> void
#
# The call placePiece(piece, board, row, col) will place piece at
# row, col in the board. It also takes care of updating the user
# interface. You don't have to deal with the user interface.
#
def placePiece(piece, board, row, col):
    global buttonFinder, imageDictionary
    board[row][col] = piece
    image = imageDictionary[piece]
    button = buttonFinder[(row, col)]
    button.config(image=image)
    button.image = image

# show : board -> void
#
# The call show(board) will show a visualization of the board in
# the user interface. The show function only updates the user
# iterface sometimes.  It's too slow to update it on every change.
# You do not have to deal with the user interface.
#
def show(board):
    global showCount
    showSkips = 1000
    showCount += 1
    if showCount % showSkips == 0:
        userInterface.update()
        userInterface.update_idletasks()
  
# Predicates for piece directions. E.g., goesUp : piece -> boolean
#
def goesUp(piece):
    return isEmpty(piece) or (piece != Wall and shapeOf(piece) in [Vertical, Q1, Q2, Circle])

# YOUR CODE HERE: You'll need functions to test for goesDown, 
#                 goesLeft, goesRight.

# See if the current piece is OK with respect to what is in the
# cell above.
#
def isAllowedAbove(above, curr):
    if isEmpty(above):
        return True # empty cells always allowed above
    if not goesUp(curr) and not goesDown(above):
        return True # tiles do not join each other; allowed
    if not goesUp(curr) and isEndPiece(above):
        return True # endpoints go every direction, but it's not required that you enter them
    if goesUp(curr) and goesDown(above) and sameColor(curr, above):
        return True # "normal" case of connecting tiles
    if DEBUG: print 'isnt allowed Above'
    return False # everything else

# YOUR CODE HERE: You'll need functions to test whether or not
#                 the current piece is OK with respect to what
#                 is below, left and right.


# YOUR CODE HERE: You'll need a function
#
#           checkCell : board * row * col -> boolean
#
#           which checks to see if the piece at board[row][col]
#           is OK with respect to it's neighbors above, below,
#           left and right.  In order for the piece to be OK,
#           it has to be compatible with -all- of it's neighbors
#           as would be tested by the isAllowedXXX functions above.


# Check to see if a given end piece is happy with its neighbors.
#
def endPieceEntryOK(board, row, col):
        entries = 0;
        left  = board[row][col - 1]
        right = board[row][col + 1]
        above = board[row - 1][col]
        below = board[row + 1][col]
        empties = 0;
        if isEmpty(left):
            empties += 1
        if isEmpty(right):
            empties += 1
        if isEmpty(above):
            empties += 1
        if isEmpty(below):
            empties += 1
            
        if not isEmpty(left) and (not isEndPiece(left)) and goesRight(left):
            entries += 1
        if not isEmpty(right) and (not isEndPiece(right)) and goesLeft(right):
            entries += 1
        if not isEmpty(above) and (not isEndPiece(above)) and goesDown(above):
            entries += 1
        if not isEmpty(below) and (not isEndPiece(below)) and goesUp(below):
            entries += 1
                                  
        if entries > 1 or (entries == 0 and empties == 0):
            return False
        return True

# YOUR CODE HERE: 
#
#              You'll need at least two functions:
#
#              endPieceEntriesOK : board -> boolean
#
#              This function should search through the board
#              and for any end piece found, it should confirm
#              that it is OK with respect to its neighbors. 
#              If so, return True. If not, return False.
#
#              You'll need another function:
#
#              tryAllPieces : board * row * col -> boolean
#
#              This function should try every playable piece in 
#              board[row][col]. (use the placePiece function, then 
#              show(board)).
#              
#              Once a piece is placed on the board, it should be
#              checked to ensure that it's compatible with it's
#              neighbors.  If it is, see if all of the end piece
#              entries are OK. If they are, then recursively try to
#              solve the board. If it succeeds, then you're done:
#              return True.
#
#              If on the other hand, the piece isn't compatible with
#              its neighbors, or any of the end pieces are not OK or
#              the remaining board was unsolvable, then move on to try
#              the next playable piece.
#
#              If none of the playable pieces lead to a solution, then
#              tryAllPieces failed. Place the Empty piece in the board
#              at row, col, show the board, and return False.

                                  
# solve the board by playing a piece in first empty cell. Use
# for-loops together with recursion to implement backtracking.
#
def solve(board):
    # YOUR CODE HERE!!!
    #
    # Find the first empty cell. If trying all of the pieces comes
    # up with a solution (i.e., tryAllPieces returns True) and all
    # of the end piece entries are OK, the success, return True.
    #
    # If trying all pieces in the empty entry failed, then restore
    # the Empty piece in the board at row, col, show the board and
    # return False.
    #
    # If there are no empty cells then you're all done, the puzzle
    # is solved, return True.


###################################################################
#
# FEARSOME CREATURES BELOW!          Cannot be unseen.
#
###################################################################
#
# initPiece(row, col, N) returns the appropriate piece for position
# row, col for the intial board. These are either empty or walls.
#
def initPiece(row, col, N):
    if row == 0 or row == N - 1 or col == 0 or col == N - 1:
        return Wall
    else:
        return Empty

# getEndPiece(n) returns the same end piece for two consecutive
# values n and n+1. Used in initialization.
#
def getEndPiece(n): return pieces[((n / 2) + 1) * nShapes - 1]


# The show starts here!
#
def start():

    global userInterface
    
    Ns = range(N + 2)  # Two extra rows and columns for walls.
    
    board = [[ initPiece(row, col, N + 2) for col in Ns] for row in Ns]

    makeUserInterface(board)
    
    userInterface.mainloop()


# makeImageDictionary : pieces -> (piece, image) dictionary
#
# This function makes a dictionary mapping pieces to images of
# those pieces.
#
def makeImageDictionary(pieces):
    dictionary = {}
    for piece in pieces:
        fileName = pieceName(piece) + '.gif'       
        dictionary[piece] = PhotoImage(file="PieceImages/" + fileName)
    dictionary[Empty] = PhotoImage(file="PieceImages/Empty.gif")
    dictionary[Wall] = PhotoImage(file="PieceImages/Wall.gif")

    return dictionary

clickCount = 0
buttonFinder = {}

def makeUserInterface(board):

    global userInterface, buttonFinder, imageDictionary
    
    def buttonClick(event):
        global clickCount
        endPiece = getEndPiece(clickCount)
        row = event.widget.row
        col = event.widget.col
        placePiece(endPiece, board, row, col)
        clickCount = clickCount + 1

    userInterface = Tk()
    userInterface.title("CS101 Flow Solver")

    imageDictionary = makeImageDictionary(pieces)

    for row in range(N + 2):
        for col in range(N + 2):
            piece = board[row][col]
            image = imageDictionary[piece]
            if isEmpty(piece):
                button = Button(userInterface, image=image)
                button.row = row
                button.col = col

                # Save the reference to this button widget so we can find
                # it later having only board coordinates row and col. No 
                # doubt there is a better way to do this ...
                #
                buttonFinder[(row, col)] = button
                
                button.grid(row=row - 1, column=col - 1)
                button.bind("<Button-1>", buttonClick)
            else:              
                button = Label(userInterface, image=image)                

            button.image = image

    solveButton = Button(userInterface, text="Solve", command=lambda : solve(board))
    solveButton.grid(row=N + 2, column=0, columnspan=2)

    testButton = Button(userInterface, text="Test Case", command=lambda : test2(board))
    testButton.grid(row=N + 2, column=3, columnspan=2)


# A couple of tester configurations.
#
def test1(board):
    placePiece(Q1, board, 3, 3)
    placePiece(Q1, board, 4, 4)
    if DEBUG: pprint.pprint(board)
    show(board)
    
def test2(board):
    p0 = getEndPiece(0)
    p1 = getEndPiece(2)
    p2 = getEndPiece(4)
    p3 = getEndPiece(6)
    placePiece(p0, board, 2, 2)
    placePiece(p0, board, 5, 3)
    placePiece(p1, board, 3, 4)
    placePiece(p1, board, 6, 4)
    placePiece(p2, board, 6, 3)
    placePiece(p2, board, 5, 4)
    placePiece(p3, board, 2, 5)
    placePiece(p3, board, 5, 5)

    show(board)

showCount = 0   
start()
